#include <suifkernel/suif_env.h>
#include <suifkernel/module_subsystem.h>
#include <suifkernel/module.h>
#include <suifkernel/command_line_parsing.h>
#include <suifpasses/passes.h>
#include <stdio.h>

#include <basicnodes/basic.h>
#include <basicnodes/basic_factory.h>

#include <suifnodes/suif.h>

#include <esuifnodes/esuif.h>

extern void est2s_parse_file(const char *, FileSetBlock *, SuifEnv *);

static String _file_name;

class Est2sPass : public FrontendPass {
public:
  Est2sPass(SuifEnv *env, const LString &name)
    : FrontendPass(env, name) {}

  ~Est2sPass() {}

  Module *clone() const;
  void initialize();

  /// Does the work: called by FrontendPass::execute()
  FileSetBlock *build_file_set_block();
};

Module *Est2sPass::clone() const
{
  return (Module*)this;
}

void Est2sPass::initialize()
{
  FrontendPass::initialize();
  _command_line->set_description("Create SUIF for an Esterel source file.");
  OptionString *_file_name_argument =
    new OptionString("Filename", &_file_name);
  _file_name_argument->set_description("Input file name");
  _command_line->add(_file_name_argument);
}

FileSetBlock *Est2sPass::build_file_set_block()
{
  printf("build_file_set_block called: filename %s\n", _file_name.c_str());

  FileSetBlock *fsb = create_file_set_block(get_suif_env());
  // FileSetBlock *fsb = get_suif_env()->get_file_set_block();
 
  suif_assert(fsb);

  est2s_parse_file( _file_name.c_str(), fsb, get_suif_env() );

  return fsb;
}

extern "C" void init_est2s(SuifEnv *suif_env)
{
  init_basicnodes(suif_env);
  init_suifnodes(suif_env);
  init_esuifnodes(suif_env);
  // Add more modules to initialize (?)
  ModuleSubSystem *ms = suif_env->get_module_subsystem();
  if (!ms->retrieve_module("est2s"))
    ms->register_module(new Est2sPass(suif_env, "est2s"));
}
