header "post_include_hpp" {
#include <suifkernel/suif_env.h>
#include <basicnodes/basic_forwarders.h>
#include <suifnodes/suif_forwarders.h>
#include <esuifnodes/esuif_forwarders.h>
#include <common/lstring.h>
#include <common/i_integer.h>
#include <common/suif_indexed_list.h>
#include <antlr/Token.hpp>
#include "LineAST.hpp"
#include <stdarg.h>
#include <utils/symbol_utils.h>
#include <typebuilder/type_builder.h>
class SuifEnv;
}

header "post_include_cpp" {
}

header {

/**********************************************************************

The c2s frontend:

For 

void main() { int a; }

FileSetBlock.external_symbol_table:

  built-in C types: a BooleanType, an unsigned IntegerType, a signed
  IntegerType, etc.  None have names.

  CProcedureType for the "void main()" function, unnamed.
  A QualifiedType pointing to the signed integer NumericType;
  ProcedureSymbol for "main", named.

FileSetBlock.file_set_symbol_table:

  empty

FileBlocks contains a single FileBlock

The FileBLock contains a single ProcedureDefinition

ProcedureDefinition.symbolTable contains the "a" variable symbol whose 
type is the QualifiedType for integer in the external_symbol_table.

The explicit_super_scope is symbol_table for the FileBlock, which is
empty.

VariableDefintions appear in FileBlock.definition_block for
global variables, not automatic variables.

Initialization of automatic variables is transformed into
StoreVariableStatements at the beginning of the procedure.

***********************************************************************/

#include <basicnodes/basic.h>
#include <basicnodes/basic_factory.h>
#include <suifnodes/suif.h>
#include <suifnodes/suif_factory.h>
#include <esuifnodes/esuif.h>
#include <esuifnodes/esuif_factory.h>

LString EsterelTreeParser::inputAnnote = "input";
LString EsterelTreeParser::outputAnnote = "output";
LString EsterelTreeParser::inputoutputAnnote = "inputoutput";
LString EsterelTreeParser::returnAnnote = "return";
LString EsterelTreeParser::sensorAnnote = "sensor";
LString EsterelTreeParser::localAnnote = "local";
LString EsterelTreeParser::integerTypeName = "integer";
LString EsterelTreeParser::booleanTypeName = "boolean";
LString EsterelTreeParser::stringTypeName = "string";
LString EsterelTreeParser::floatTypeName = "float";
LString EsterelTreeParser::doubleTypeName = "double";
LString EsterelTreeParser::lineAnnote = "line";
LString EsterelTreeParser::constantQualifier = "const";
LString EsterelTreeParser::tickSignalName = "tick";
LString EsterelTreeParser::voidTypeName = "void";

#define OPSTRINGS \
opstring(add); \
opstring(subtract); \
opstring(multiply); \
opstring(divide); \
opstring(remainder); \
opstring(is_equal_to); \
opstring(is_not_equal_to); \
opstring(is_less_than); \
opstring(is_less_than_or_equal_to); \
opstring(is_greater_than); \
opstring(is_greater_than_or_equal_to); \
opstring(logical_and); \
opstring(logical_or); \
opstring(negate); \
opstring(logical_not)

#define opstring(x) LString EsterelTreeParser::x ## _op = #x
OPSTRINGS;
#undef opstring

QualifiedType *EsterelTreeParser::lookup_builtin_type(LString name)
{
  indexed_list<LString, QualifiedType*>::const_iterator iter =
    builtin_types.find(name);
  if (iter == builtin_types.end()) return 0;
  return (*iter).second; 
}

BuiltinFunctionSymbol *EsterelTreeParser::get_builtin_function(Context *c,
							       LString name)
{
  suif_assert(c);
  suif_assert(c->module);
  SymbolTableObject *sto =
    c->module->get_builtin_symbols()->lookup_lookup_table(name);
  if (sto) {
    return to<BuiltinFunctionSymbol>(sto);
  }
  BuiltinFunctionSymbol *bfs = create_builtin_function_symbol(suif_env, 0);
  c->module->get_builtin_symbols()->add_symbol(name, bfs);
  return bfs;
}

QualifiedType* EsterelTreeParser::get_const_type(Context *c,
						 DataType *type)
{
  for ( Iter<SymbolTableObject *> iter =
	  c->module->get_symbol_table()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    SymbolTableObject *sto = iter.current();
    if ( is_a<QualifiedType>(sto) ) {
      QualifiedType *qt = to<QualifiedType>(sto);
      if (qt->get_base_type() == type &&
	  qt->has_qualification_member(constantQualifier))
        return qt;
    }
  }
  QualifiedType *qt = create_qualified_type(suif_env, type);
  qt->append_qualification(constantQualifier);
  qt->set_name(type->get_name());
  c->variables->append_symbol_table_object(qt);
  return qt;
}

VariableSymbol *EsterelTreeParser::find_variable(Context *c,
						 RefLineAST symbol)
{
  LString name(symbol->getText().c_str());
  for (SymbolTable *st = c->variables ;
       st ; st = st->get_explicit_super_scope()) {
    SymbolTableObject * sto = st->lookup_lookup_table(name);
    if (sto) return to<VariableSymbol>(sto);
  }
  report_error(symbol, "undeclared variable `%s'", name.c_str());

  // Create a fictious integer variable as a placeholder

  VariableSymbol *vs =
    create_variable_symbol(suif_env, info_block->get_q_integer_type(), name);
  return vs;
}

TrapSymbol *EsterelTreeParser::find_trap(Context *c,
					 RefLineAST symbol)
{
  LString name(symbol->getText().c_str());
  for (SymbolTable *st = c->traps ;
       st ; st = st->get_explicit_super_scope()) {
    SymbolTableObject * sto = st->lookup_lookup_table(name);
    if (sto) return to<TrapSymbol>(sto);
  }
  report_error(symbol, "undeclared trap `%s'", name.c_str());

  // Create a fictious pure exception as a placeholder

  VariableSymbol *var = new_anonymous_variable(suif_env, c->traps, info_block->get_q_void_type());
  TrapSymbol *ts = create_trap_symbol(suif_env, info_block->get_q_void_type(), var, 0, name);
  return ts;
}

SignalSymbol *EsterelTreeParser::find_signal(Context *c,
					     RefLineAST symbol)
{
  LString name(symbol->getText().c_str());
  for (SymbolTable *st = c->signals ;
       st ; st = st->get_explicit_super_scope()) {
    SymbolTableObject * sto = st->lookup_lookup_table(name);
    if (sto) return to<SignalSymbol>(sto);
  }
  report_error(symbol, "undeclared signal `%s'", name.c_str());

  // Create a fictious pure signal as a placeholder

  VariableSymbol *var =
    new_anonymous_variable(suif_env, c->signals, info_block->get_q_boolean_type());
  SignalSymbol *ss =
    create_signal_symbol(suif_env, info_block->get_q_void_type(), inputAnnote, var, 0, 0, name);
  return ss;
}

Expression* EsterelTreeParser::build_boolean_binary(RefLineAST token,
						    LString op,
						    Expression *e1,
						    Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  Expression *e =
    create_binary_expression(suif_env, e1->get_result_type(), op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if ( dt1 != dt2 || !is_a<BooleanType>(dt1) ) {
    report_error(token, "arguments of `%s' must both be boolean", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression* EsterelTreeParser::build_numeric_binary(RefLineAST token,
						    LString op,
						    Expression *e1,
						    Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  Expression *e =
    create_binary_expression(suif_env, e1->get_result_type(), op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if ( dt1 != dt2 || !(is_a<IntegerType>(dt1) ||
		       is_a<FloatingPointType>(dt1)) ) {
    report_error(token, "arguments of `%s' must be numeric and match", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression* EsterelTreeParser::build_equality_binary(RefLineAST token,
						     LString op,
						     Expression *e1,
						     Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  Expression *e =
    create_binary_expression(suif_env, info_block->get_boolean_type(),
			     op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if ( dt1 != dt2 ) {
    report_error(token, "type of arguments of `%s' must match", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression* EsterelTreeParser::build_comparison_binary(RefLineAST token,
						       LString op,
						       Expression *e1,
						       Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  Expression *e =
    create_binary_expression(suif_env, info_block->get_boolean_type(),
			     op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if (dt1 != dt2 || !(is_a<IntegerType>(dt1) || is_a<FloatingPointType>(dt1))){
    report_error(token, "arguments of `%s' must be numeric and match", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression *EsterelTreeParser::build_boolean_unary(RefLineAST token,
						   LString op,
						   Expression *e1)
{
  suif_assert(e1);
  Expression *e =
    create_unary_expression(suif_env, info_block->get_boolean_type(), op, e1);

  DataType *dt1 = e1->get_result_type();
  suif_assert(dt1);
  if (!is_a<BooleanType>(dt1)) {
    report_error(token, "arguments of `%s' must be boolean", op.c_str());
    report_warning(token, "argument type is `%s'", dt1->get_name().c_str());
  }
  return e;
}

Expression *EsterelTreeParser::build_numeric_unary(RefLineAST token,
						   LString op,
						   Expression *e1)
{
  suif_assert(e1);
  Expression *e =
    create_unary_expression(suif_env, e1->get_result_type(), op, e1);

  DataType *dt1 = e1->get_result_type();
  suif_assert(dt1);
  if (is_a<BooleanType>(dt1)) {
    report_error(token, "arguments of `%s' must be boolean", op.c_str());
    report_warning(token, "argument type is `%s'", dt1->get_name().c_str());
  }
  return e;
}

void EsterelTreeParser::create_constant(Context *c,
					RefLineAST id,
					Expression *e,
					RefLineAST expr,
					QualifiedType *qt)
{
  LString name(id->getText().c_str());
  SymbolTableObject *sto = c->variables->lookup_lookup_table(name);
  if (sto) {
    report_error(id, "redeclaration of constant `%s'", name.c_str());
    report_warning(sto, "`%s' previously declared here", name.c_str());
  }

  QualifiedType *qtc = get_const_type(c, qt->get_base_type());

  // Add a VariableSymbol to md->symbol_table
  //   QualifiedType must be a variant with const
  //   VariableDefinition
  //
  // Add VariableDefinition to md->definition_block->variable_definitions
  //   VariableSymbol points to above
  //   ValueBlock is an ExpressionValueBlock      

  VariableSymbol *vs = create_variable_symbol(suif_env, qtc);
  attach_token_linenumber(vs, id);
  c->variables->add_symbol(name, vs);

  ValueBlock *vb = e ? create_expression_value_block(suif_env, e) : 0;
  VariableDefinition *vd = create_variable_definition(suif_env, vs, 0, vb);
  c->module->get_definition_block()->append_variable_definition(vd);

  vs->set_definition(vd);

  if (e) {
    if (e->get_result_type() != qt->get_base_type()) {
      report_error(id, "initializer does not match type of its constant");
      report_warning(id, "type of constant is `%s'", qt->get_name().c_str());
      report_warning(expr, "type of initializer is `%s'",
		     e->get_result_type()->get_name().c_str());
    }
  }
}

void EsterelTreeParser::create_local_variable(Context *c,
					      VariableStatement *vst,
					      RefLineAST id,
					      Expression *e,
					      RefLineAST expr,
					      QualifiedType *qt)
{
  LString name(id->getText().c_str());
  suif_assert(c->variables);
  SymbolTableObject *sto = c->variables->lookup_lookup_table(name);
  if (sto) {
    report_error(id, "redefinition of local variable `%s'",
		 name.c_str());
    report_warning(sto, "`%s' previously defined here", name.c_str());
  }
  VariableSymbol *vsy = create_variable_symbol(suif_env, qt);
  attach_token_linenumber(vsy, id);
  c->variables->add_symbol(name, vsy);

  ValueBlock *vb = e ? create_expression_value_block(suif_env, e) : 0;
  VariableDefinition *vd =
    create_variable_definition(suif_env, vsy, 0, vb);
  vst->get_definition_block()->append_variable_definition(vd);
  vsy->set_definition(vd);

  if (e) {
    if (e->get_result_type() != qt->get_base_type()) {
      report_error(id, "initializer does not match type of its variable");
      report_warning(id, "type of variable is `%s'", qt->get_name().c_str());
      report_warning(expr, "type of initializer is `%s'",
		     e->get_result_type()->get_name().c_str());
    }
  }
}

/**
   Create a new variable by adding a unique numeric suffix to
   the given base name.
 */
VariableSymbol *EsterelTreeParser::new_variable(SuifEnv *env,
						SymbolTable *st,
						QualifiedType *qt,
						String base)
{
  String name;
  for (IInteger i = 0 ; ; i++) {
    name = base + i.to_String();
    if (!LString::exists(name)) {
      VariableSymbol *result = create_variable_symbol(env, qt);
      st->add_symbol(name, result);
      return result;
    }
  }
}

void EsterelTreeParser::attach_token_linenumber(AnnotableObject *ao,
						RefLineAST rt)
{
  IntegerBrick *ib = create_integer_brick(suif_env,rt->getLine());
  StringBrick *sb = create_string_brick(suif_env,current_filename);
  BrickAnnote *ba = create_brick_annote(suif_env,lineAnnote);
  ba->append_brick(ib);
  ba->append_brick(sb);
  ao->append_annote(ba);
}

void EsterelTreeParser::report_warning( AnnotableObject* ao,
					const char* format, ... )
{
  report_fileline(ao);
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
}


void EsterelTreeParser::report_error( AnnotableObject* ao,
				      const char* format, ... )
{
  report_fileline(ao);
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
  num_errors++;
}

void EsterelTreeParser::report_error( RefLineAST ast,
				      const char* format, ... )
{
  fprintf( stderr,  "%s:%d: ", current_filename.c_str(),
	                             ast->getLine() );
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
  num_errors++;
}

void EsterelTreeParser::report_warning( RefLineAST ast,
				      const char* format, ... )
{
  fprintf( stderr,  "%s:%d: ", current_filename.c_str(),
	                             ast->getLine() );
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
}

void EsterelTreeParser::report_fileline( AnnotableObject* ao )
{
  Annote * line = ao->lookup_annote_by_name(lineAnnote);
  if (line) {
    BrickAnnote *ba = to<BrickAnnote>(line);
    suif_assert( ba->get_brick_count() == 2 );
    IntegerBrick *ib = to<IntegerBrick>(ba->get_brick(0));
    StringBrick *sb = to<StringBrick>(ba->get_brick(1));
    fprintf( stderr, "%s:%d: ", sb->get_value().c_str(),
                                ib->get_value().c_int() );
  }
}

}

options {
  language="Cpp";
}

/**
 *
 * Tree walker for Esterel AST that builds the SUIF representation
 *
 * Author: Stephen A. Edwards
 *         sedwards@cs.columbia.edu
 *
 *
 * Change Log
 */
class EsterelTreeParser extends TreeParser;

options {
  importVocab = Esterel;
  ASTLabelType = "RefLineAST";	// Expect AST nodes with line numbers
  //  defaultErrorHandler=false;	// Disable default error handling for
}

{
  SuifEnv *suif_env;
  LString current_filename;
  int num_errors;

  static LString inputAnnote;
  static LString outputAnnote;
  static LString inputoutputAnnote;
  static LString returnAnnote;
  static LString sensorAnnote;
  static LString localAnnote;
  static LString integerTypeName;
  static LString booleanTypeName;
  static LString stringTypeName;
  static LString floatTypeName;
  static LString doubleTypeName;
  static LString lineAnnote;
  static LString constantQualifier;
  static LString tickSignalName;
  static LString voidTypeName;

  static LString add_op;
  static LString subtract_op;
  static LString multiply_op;
  static LString divide_op;
  static LString remainder_op;
  static LString is_equal_to_op;
  static LString is_not_equal_to_op;
  static LString is_less_than_op;
  static LString is_less_than_or_equal_to_op;
  static LString is_greater_than_op;
  static LString is_greater_than_or_equal_to_op;
  static LString logical_and_op;
  static LString logical_or_op;
  static LString negate_op;
  static LString logical_not_op;

  struct Context {
    ModuleDefinition *module;
    SymbolTable *types;
    SymbolTable *functions;
    SymbolTable *procedures;
    SymbolTable *tasks;
    SymbolTable *variables;
    SymbolTable *traps;
    SymbolTable *signals;
    SymbolTable *signal_vars;
    DefinitionBlock *definitions;  
  };

public:
  void set_suif_env(SuifEnv *env) { suif_env = env; }

  EsterelInformationBlock *info_block;

  /// Symbol table for built-in types
  indexed_list<LString, QualifiedType*> builtin_types;
  QualifiedType *lookup_builtin_type(LString);

  /// Find or create a built-in function with the given name
  BuiltinFunctionSymbol *get_builtin_function(Context *, LString);

  /// Find or create a constant type with the given base type
  QualifiedType *get_const_type(Context *, DataType *);

  /// Locate a given variable symbol or produce an error
  VariableSymbol *find_variable(Context *, RefLineAST);

  /// Locate a given exception symbol or produce an error
  TrapSymbol *find_trap(Context *, RefLineAST);

  /// Locate a given signal symbol or produce an error
  SignalSymbol *find_signal(Context *, RefLineAST);

  /// Construct and typecheck a Boolean (and, or) binary operator
  Expression *build_boolean_binary(RefLineAST, LString, Expression*, Expression*);

  /// Construct and typecheck a Numeric (+, *) binary operator
  Expression *build_numeric_binary(RefLineAST, LString, Expression*, Expression*);

  /// Construct and typecheck a equality (=, <>) binary operator
  Expression *build_equality_binary(RefLineAST, LString, Expression*, Expression*);

  /// Construct and typecheck a comparison (<, <=, etc.) binary operator
  Expression *build_comparison_binary(RefLineAST, LString, Expression*, Expression*);

  /// Construct and typecheck a boolean (not) unary operator
  Expression *build_boolean_unary(RefLineAST, LString, Expression*);

  /// Construct and typecheck a numeric (-) unary operator
  Expression *build_numeric_unary(RefLineAST, LString, Expression*);

  /// Create a constant in the module
  void create_constant(Context *, RefLineAST, Expression *,
		       RefLineAST, QualifiedType *);

  /// Create a local variable in a variable statement
  void create_local_variable(Context *, VariableStatement *,  RefLineAST,
			     Expression *, RefLineAST, QualifiedType *);

  /// Create a new variable by adding a unique numeric suffix
  VariableSymbol *new_variable(SuifEnv *, SymbolTable *, QualifiedType *,
			       String);

  /// Attach line number information to an AnnotableObject
  void attach_token_linenumber(AnnotableObject *, RefLineAST);


  /// Report a warning
  void report_warning(AnnotableObject *, const char *, ... );
  void report_warning(RefLineAST, const char *, ... );

  /// Report an error
  void report_error(AnnotableObject *, const char *, ... );
  void report_error(RefLineAST, const char *, ... );

  /// Return the number of errors
  int get_num_errors() const { return num_errors; }

  /// Report file:line: for an annotated object.
  void report_fileline(AnnotableObject*);
  
}

file [FileSetBlock* fsb, LString filename]
  {
    FileBlock *fb = create_file_block(suif_env, filename);
    fsb->append_file_block(fb);
    current_filename = filename;
    num_errors = 0;
    info_block =
      to<EsterelInformationBlock>(get_global_information_block(suif_env,
			          EsterelInformationBlock::get_class_name()));
    suif_assert(info_block); // Should have been attached by the pass
  }
  : (module[fb])+
  ;
/*
exception
catch [ANTLR_USE_NAMESPACE(antlr)NoViableAltException& ex] {
  report_error(static_cast<RefLineAST>(ex.node),
	       "Unexpected AST node `%s'",
	       ex.toString().c_str());
}
*/

/*
 * Line number information:
 * A BrickAnnote called "line" contains an IntegerBrick with the line number
 * and a StringBrick with the filename.
 */

module [FileBlock* fb]
  : #( "module" moduleName:ID
       {
	 // Check for a duplicate module definition

	 LString name(moduleName->getText().c_str());
         SymbolTableObject* sto =
           fb->get_symbol_table()->lookup_lookup_table(name);
         if ( sto ) {	   
	   report_error(moduleName, "redefinition of module `%s'",
                                      name.c_str() );
           report_warning(sto, "`%s' previously defined here", name.c_str());
         }

	 // Create the module symbol and attach it to
	 // the file set block's external symbol table, in effect marking it
	 // as non-static

         ModuleSymbol *ms = create_module_symbol(suif_env, (ProcedureType*) 0);
	 attach_token_linenumber(ms, moduleName);
	 suif_env->get_file_set_block()->get_external_symbol_table()->add_symbol(name, ms);

         // Create a new module and add it to the FileBlock

         ModuleDefinition *md = create_module_definition(suif_env);
         md->set_procedure_symbol(ms);
	 attach_token_linenumber(md, moduleName);
         fb->get_definition_block()->append_procedure_definition(md);

	 Context c;
	 c.module = md;
	 c.types = md->get_type_symbols();
	 c.functions = md->get_function_symbols();
	 c.procedures = md->get_procedure_symbols();
	 c.tasks = md->get_task_symbols();
	 c.variables = md->get_symbol_table();
	 c.traps = 0;	// No traps at the outer level
	 c.signals = md->get_signal_symbols();
	 // Variables for interface signals end up outside the reaction func.
	 c.signal_vars = fb->get_symbol_table();
	 c.definitions = md->get_definition_block();	 

         // Add the built-in types to the "builtin_types" symbol table
	 // that is not part of the SUIF file written out

	 TypeBuilder *tb = get_type_builder(suif_env);

	 builtin_types.push_back(integerTypeName,
				 info_block->get_q_integer_type());
	 builtin_types.push_back(booleanTypeName,
				 info_block->get_q_boolean_type());
	 builtin_types.push_back(stringTypeName,
		    tb->get_qualified_type(info_block->get_string_type()));
	 builtin_types.push_back(floatTypeName,
		    tb->get_qualified_type(info_block->get_float_type()));
	 builtin_types.push_back(doubleTypeName,
		    tb->get_qualified_type(info_block->get_double_type()));

	 // Built-in input signal "tick"

	 VariableSymbol *tick_var =
	   new_variable(suif_env, c.signal_vars,
			info_block->get_q_boolean_type(), "tick_");
	 VariableDefinition *tick_def =
	   create_variable_definition(suif_env, tick_var, 0, 0, true);
	 tick_var->set_definition(tick_def);	 
	 c.definitions->append_variable_definition(tick_def);
	 SignalSymbol *ss =
	   create_signal_symbol(suif_env, info_block->get_q_void_type(), inputAnnote, tick_var, 0, 0);
	 c.signals->add_symbol(tickSignalName, ss);
       }
       declarations[&c]
       {
         Statement *sta;
       }
       sta=statement[&c]
       {
         md->set_body(sta); 
       }
     )
  ;

declarations [Context *c]
  : #( DECLS
       ( #( "type" ( typeDecl[c] )+ )
       | #( "constant" (constantDecl[c])+ )
       | #( "input" ( signalDecl[c, inputAnnote, true] )+ )
       | #( "output" ( signalDecl[c, outputAnnote, true] )+ )
       | #( "inputoutput" ( signalDecl[c, inputoutputAnnote, true] )+ )
       | #( "return" ( signalDecl[c, returnAnnote, true] )+ )
       | #( "function" ( functionDecl[c] )+ )
       | #( "procedure" ( procedureDecl[c] )+ )
       | #( "task" ( taskDecl[c] )+ )
       | #( "sensor" ( signalDecl[c, sensorAnnote, true] )+ )
       | #( "relation" ( relationDecl[c] )+ )
       )*
     )
  ;

typeDecl [Context *c]
  : typeName:ID {
       LString name(typeName->getText().c_str());
       SymbolTableObject *sto =
          c->types->lookup_lookup_table(name);
       if (sto) {     
           report_error(typeName, "redeclaration of type `%s'", name.c_str());
           report_warning(sto, "`%s' previously declared here", name.c_str());
       }
       if (lookup_builtin_type(name)) {
	 report_error(typeName, "redeclaration of built-in type `%s'", name.c_str());
       }
       ExternalType *et = create_external_type(suif_env,0,0,name);
       c->types->append_symbol_table_object(et);
       QualifiedType *qt = create_qualified_type(suif_env, et);
       attach_token_linenumber(qt, typeName);
       c->types->add_symbol(name, qt);
    }
  ;

constantDecl [Context *c]
  { Expression *e = 0;
    QualifiedType *qt; }
  : #(CDECL
      ( constantName:ID ( #(EQUALS e=expr:expression[c]) )?
        qt=typeToken:type[c]
        { create_constant(c, constantName, e, expr, qt); }
      | #(ids:COMMA (ID)+) qt=type[c]
        {
	  for ( RefLineAST id = #ids->getFirstChild() ; id ;
		id=id->getNextSibling() )
	    create_constant(c, id, 0, 0, qt);
	}
      )
    )
  ;

signalDecl [Context *c, LString direction, bool isGlobal]
  : #(SDECL signalName:ID
       {
         LString name(signalName->getText().c_str());
         SymbolTableObject *sto =
            c->signals->lookup_lookup_table(name);

         if (sto) {
           report_error(signalName, "redeclaration of `%s'", name.c_str());
           report_warning(sto, "`%s' previously declared here", name.c_str());
         }
         Expression *e = 0;
       }
       ( #(COLEQUALS e=expr:expression[c]) )?
       {
         QualifiedType *qt = info_block->get_q_void_type();
	 EsterelFunctionSymbol *efs = 0;
       }
       (qt=typeToken:type[c]
         ( func:ID
           {
	     LString name = func->getText().c_str();
	     SymbolTableObject *sto = c->functions->lookup_lookup_table(name);
	     if (!sto)
	       report_error(func, "undeclared function `%s'", name.c_str());
	     else
	       efs = to<EsterelFunctionSymbol>(sto);
	   }
         | pcf:predefinedCombineFunction
           { efs = get_builtin_function(c, pcf->getText().c_str()); }
         )?
       )?
       {
	 // Create a variable for the signal's presence if it's not a sensor

	 VariableSymbol *presence_var;
	 if ( direction != sensorAnnote ) {
	   presence_var =
	     new_variable(suif_env, c->signal_vars,
			  info_block->get_q_boolean_type(),
			  name);
	   VariableDefinition *vd =
	     create_variable_definition(suif_env, presence_var, 0, 0, isGlobal);
	   c->definitions->append_variable_definition(vd);
	 } else {
	   presence_var = 0;
	 }

	 // Create a variable for the signal's value if it's not pure (void)

	 VariableSymbol *value_var;
	 if (qt != info_block->get_q_void_type()) {
	   value_var = new_variable(suif_env, c->signal_vars, qt,
				    name + "_val");
	     ValueBlock *vb =
	       e ? create_expression_value_block(suif_env, e) : 0;
	     VariableDefinition *vd =
	       create_variable_definition(suif_env, value_var, 0, vb, isGlobal);
	     c->definitions->append_variable_definition(vd);
	     value_var->set_definition(vd);
	 } else {
	   value_var = 0;
	 }

	 // Create the signal symbol

         SignalSymbol *ss =
	   create_signal_symbol(suif_env, qt, direction, presence_var,
				value_var, efs);
         attach_token_linenumber(ss, signalName);
         c->signals->add_symbol(name,ss);

	 if (e) {
	   suif_assert(qt->get_base_type());
	   if (e->get_result_type() != qt->get_base_type()) {
	     report_error(typeToken,
			  "initializer does not match type of signal");
	     report_warning(typeToken, "type of signal is `%s'",
			    qt->get_name().c_str());
	     report_warning(expr, "type of initializer is `%s'",
			    e->get_result_type()->get_name().c_str());
	   }
	 }
       }
     )
  ;

trapDecl [Context *c]
  : #(TDECL trapName:ID
       {
         LString name(trapName->getText().c_str());
         SymbolTableObject *sto =
            c->traps->lookup_lookup_table(name);

         if (sto) {
           report_error(trapName, "redeclaration of `%s'", name.c_str());
           report_warning(sto, "`%s' previously declared here", name.c_str());
         }
         Expression *e = 0;
       }
       ( #(COLEQUALS e=expr:expression[c]) )?
       {
         QualifiedType *qt = info_block->get_q_void_type();
       }
       ( qt=typeToken:type[c] )?
       {
	 // Create a variable for the trap's presence

	 VariableSymbol *presence_var;
	 presence_var =
	   new_variable(suif_env, c->traps,
			info_block->get_q_boolean_type(), name);
	 VariableDefinition *vd =
	   create_variable_definition(suif_env, presence_var, 0, 0);
	 c->definitions->append_variable_definition(vd);

	 // Create a variable for the traps's value if it's not pure (void)

	 VariableSymbol *value_var;
	 if (qt != info_block->get_q_void_type()) {
	   value_var = new_variable(suif_env, c->traps, qt,
				    name + "_val");
	     ValueBlock *vb =
	       e ? create_expression_value_block(suif_env, e) : 0;
	     VariableDefinition *vd =
	       create_variable_definition(suif_env, value_var, 0, vb);
	     c->definitions->append_variable_definition(vd);
	     value_var->set_definition(vd);
	 } else {
	   value_var = 0;
	 }

	 // Create the trap symbol

         TrapSymbol *ts =
	   create_trap_symbol(suif_env, qt, presence_var, value_var);
         attach_token_linenumber(ts, trapName);
         c->traps->add_symbol(name,ts);

	 if (e) {
	   suif_assert(qt->get_base_type());
	   if (e->get_result_type() != qt->get_base_type()) {
	     report_error(typeToken,
			  "initializer does not match type of trap");
	     report_warning(typeToken, "type of signal is `%s'",
			    qt->get_name().c_str());
	     report_warning(expr, "type of initializer is `%s'",
			    e->get_result_type()->get_name().c_str());
	   }
	 }
       }
     )
  ;

type [Context *c] returns [QualifiedType* qt]
  :
    type:ID {
      LString name(type->getText().c_str());
      SymbolTableObject *sto = c->types->lookup_lookup_table(name);
      if (sto) {
        qt = to<QualifiedType>(sto);
      } else {
	qt = lookup_builtin_type(name);
	if (!qt) {
	  report_error(type, "unknown type `%s'", name.c_str());
	  // Define the type to avoid future errors
	  ExternalType *et = create_external_type(suif_env,0,0,name);
	  c->types->append_symbol_table_object(et);
	  qt = create_qualified_type(suif_env, et);
	  attach_token_linenumber(et, type);
	  c->types->add_symbol(name, qt);
	}
      }
    }
  ;

predefinedCombineFunction
  : PLUS
  | STAR
  | "or"
  | "and"
  ;

functionDecl [Context *c]
  : #( FDECL functionName:ID
    {
      LString name(functionName->getText().c_str());
      SymbolTableObject *sto =
        c->functions->lookup_lookup_table(name);
      if (sto) {
        report_error(functionName,
		     "redeclaration of function `%s'", name.c_str());
        report_warning(sto, "`%s' previously declared here", name.c_str());
      }
      EsterelFunctionType *eft = create_esterel_function_type(suif_env);
      c->functions->append_symbol_table_object(eft);
      EsterelFunctionSymbol *efs =
        create_esterel_function_symbol(suif_env, eft);
      attach_token_linenumber(efs, functionName);
      c->functions->add_symbol(name,efs);
      QualifiedType *qt, *rt;
    }
    #(TYPES (qt=type[c] { eft->append_argument(qt); } )* )
    rt=type[c]
    {
      eft->set_result(rt);
    }
    )
  ;

procedureDecl [Context *c]
  : #( PDECL procedureName:ID
    {
      LString name(procedureName->getText().c_str());
      SymbolTableObject *sto =
        c->procedures->lookup_lookup_table(name);
      if (sto) {
        report_error(procedureName, "redeclaration of procedure `%s'", name.c_str());
        report_warning(sto, "`%s' previously declared here", name.c_str());
      }
      EsterelProcedureType *ept = create_esterel_procedure_type(suif_env);
      c->procedures->append_symbol_table_object(ept);
      EsterelProcedureSymbol *eps =
         create_esterel_procedure_symbol(suif_env, ept);
      attach_token_linenumber(eps, procedureName);
      c->procedures->add_symbol(name,eps);
      QualifiedType *qt;
    }
    #(TYPES (qt=type[c] { ept->append_reference_argument(qt); } )* )
    #(TYPES (qt=type[c] { ept->append_value_argument(qt); } )* )  
    )
  ;

taskDecl [Context *c]
  : #( TADECL taskName:ID
    {
      LString name(taskName->getText().c_str());
      SymbolTableObject *sto =
        c->tasks->lookup_lookup_table(name);
      if (sto) {
        report_error(taskName, "redeclaration of task `%s'", name.c_str());
        report_warning(sto, "`%s' previously declared here", name.c_str());
      }
      EsterelProcedureType *ept = create_esterel_procedure_type(suif_env);
      c->tasks->append_symbol_table_object(ept);
      EsterelProcedureSymbol *eps =
         create_esterel_procedure_symbol(suif_env, ept);
      attach_token_linenumber(eps, taskName);
      c->tasks->add_symbol(name,eps);
      QualifiedType *qt;
    }
    #(TYPES (qt=type[c] { ept->append_reference_argument(qt); } )* )
    #(TYPES (qt=type[c] { ept->append_value_argument(qt); } )* )  
    )
  ;

relationDecl [Context *c]
  : #( IMPLIES sig1:ID sig2:ID
      {
	SignalSymbol *ss1 = find_signal(c, sig1);
	SignalSymbol *ss2 = find_signal(c, sig2);
	ImpliesRelation *ir = create_implies_relation(suif_env, ss1, ss2);
	attach_token_linenumber(ir, #relationDecl);
	c->module->append_relation(ir);
      }
     )
  | #( POUND
      {
	ExclusionRelation *er = create_exclusion_relation(suif_env);
	attach_token_linenumber(er, #relationDecl);
	c->module->append_relation(er);
      }
      (sig:ID
        {
	  SignalSymbol *ss = find_signal(c, sig);
	  er->append_signal(ss);
	}
      )+
     )    
  ; 

// ModuleDefinition contains the symbol table for functions, procedures, tasks.
// vs, ts, and ss are the innermost scopes for variables/constants,
// traps, and signals

statement [Context *c] returns [Statement* st]
  : ( #( PARALLEL
         { ParallelStatementList *sl =
	     create_parallel_statement_list(suif_env);
         }
         (st=statement[c] { sl->append_thread(create_thread_statement(suif_env, st)); } )+
       ) { st = sl; }

    | #( SEQUENCE { StatementList *sl = create_statement_list(suif_env); }
         (st=statement[c] { sl->append_statement(st); } )+
       ) { st = sl; }

    | "nothing" { st = create_nothing_statement(suif_env); }

    | "pause" { st = create_pause_statement(suif_env); }

    | "halt" { st = create_halt_statement(suif_env); }

    | #( "emit"
	 esig:ID
         {
           SignalSymbol *ss = find_signal(c, esig);
           Expression * e = 0;
         }
         ( e=expression[c] )?
         {
           EmitStatement *es = create_emit_statement( suif_env, ss, e );
	   // Note: A valued signal does not demand a value, but
	   // a pure signal may not have one
           if ( e && ss->get_type()->get_base_type() ==
		     info_block->get_void_type() ) {
	     report_error(esig,
			  "illegal emission of a value on pure signal `%s'",
			  esig->getText().c_str());
	   }
           st = es;
         }
      )

    | #( "sustain"
	 ssig:ID
         {
           SignalSymbol *ss = find_signal(c, ssig);
           Expression * e = 0;
         }
         ( e=expression[c] )?
         {
           EmitStatement *es = create_sustain_statement( suif_env, ss, e );
           if ( e && ss->get_type()->get_base_type() ==
		     info_block->get_void_type() ) {
	     report_error(ssig, "illegal sustain of a value on pure signal `%s'",
			  ssig->getText().c_str());
	   }
           st = es;
         }
      )

    | { Expression *e; }
      #( COLEQUALS var:ID e=expression[c] )
      { 
        VariableSymbol *vs = find_variable(c, var);
        st = create_store_variable_statement(suif_env, vs, e);
        DataType *dt1 = vs->get_type()->get_base_type();
        DataType *dt2 = e->get_result_type();
        if ( dt1 != dt2 ) {
          report_error(#statement,
	   "type mismatch: expression type `%s' must match variable type `%s'",
           dt2->get_name().c_str(), dt1->get_name().c_str());
        }
      }

    | #( "call" proc:ID
         {
           LString name = proc->getText().c_str();
	   suif_assert(c->procedures);
	   SymbolTableObject *sto = c->procedures->lookup_lookup_table(name);
	   if (!sto) {
	     report_error(proc, "undeclared procedure `%s'", name.c_str());
	     // Create a nothing statement placeholder
	     st = create_nothing_statement(suif_env);
	     break;
	   }
	   suif_assert(sto);
	   EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	   suif_assert(eps);
	   EsterelProcedureType *ept = eps->get_type();
	   suif_assert(ept);
	   ProcedureCallStatement *pcs =
	     create_procedure_call_statement(suif_env, eps);
         }
	 procArgs[c, pcs]
         { st = pcs; }
      )

    | #("present"
      {
        PresentStatement *ps = create_present_statement(suif_env);      
      }
      ( #(ca:CASE
          { Expression *e; Statement *s = 0; }
          e=sigExpression[c] (s=statement[c])?
          {
	    PredicatedStatement *ps1 =
	      create_predicated_statement(suif_env, s, e);
            attach_token_linenumber(ps1, ca);
            ps->append_case(ps1);
          }
        ) )*     
      ( #("else"
          { Statement *s = 0; }
          s=statement[c]
          { ps->set_default_case(s); }
        )
      )?
      { st = ps; }
      )
             
    | #( "if"
        {
          EsterelIfStatement *is = create_esterel_if_statement(suif_env);
	  Expression *e;
          PredicatedStatement *ps;
        }
        e=ife:expression[c]
        { Statement *s = 0; }
        ( #("then" s=statement[c] ) )?
 	{
	  ps = create_predicated_statement(suif_env, s, e); 
	  attach_token_linenumber(ps, ife);
	  is->append_case(ps);
	}
	( #(elif:"elsif" e=expression[c] s=statement[c] )
	    {
	      ps = create_predicated_statement(suif_env, s, e);
	      attach_token_linenumber(ps, elif);
	      is->append_case(ps);
	    }
	)*
        ( #("else" s=statement[c] ) { is->set_default_case(s); } )?
        { st = is; }
      )

    | #( "loop"
        { Statement *s; Expression *e; }
        s=statement[c]
        (
           /* nothing */ { st = create_loop_statement(suif_env, s); }
	 | e=delayExpression[c]
  	   { st = create_loop_each_statement(suif_env, s, e); }
        )
      )

    | #("repeat"
        {
	  bool positive = false;
          Expression *e;
        }
        ("positive" { positive=true; } )?
        e=expr:expression[c]
        {
          if ( e->get_result_type() != info_block->get_integer_type() ) {
	    report_error( expr,
			  "type of repeat count is `%s' expecting integer",
			  e->get_result_type()->get_name().c_str());
          }
	  Statement *s;
        }
        s=statement[c]
        {
	  st = create_repeat_statement(suif_env, s, e, positive); 
        }       
      )

    | #("abort"
        { bool isWeak = false; Statement *s; }
        ("weak" { isWeak = true; })?
        s=statement[c]
        {
	  AbortStatement *as = create_abort_statement(suif_env, s, 0, isWeak);
	}
        ( #(ca1:CASE { Expression *e; s = 0; }
            e=delayExpression[c] (s=statement[c])?
            {
              PredicatedStatement *ps =
		create_predicated_statement(suif_env, s, e);
	      attach_token_linenumber(ps, ca1);
	      as->append_case(ps);
	    }
	) )+
        { st = as; }
      )

    | #("await"
        { AwaitStatement *as = create_await_statement(suif_env); }
        ( #(ca2:CASE { Expression *e; Statement *s = 0; }
            e=delayExpression[c] (s=statement[c])?
            {
              PredicatedStatement *ps =
		create_predicated_statement(suif_env, s, e);
	      attach_token_linenumber(ps, ca2);
	      as->append_case(ps);
	    }
	) )+
        { st = as; }
      )

    | #("every" { Expression *e; Statement *s; }
        e=delayExpression[c] s=statement[c]
        { st = create_every_statement(suif_env, s, e); }
      )

    | #("suspend" { Expression *e; Statement *s; }
         s=statement[c] e=delayExpression[c]
        { st = create_suspend_statement(suif_env, s, e); }
      )

// FIXME: Complain if a single-exception trap uses an expression
// FIXME: Warn if a single-exception trap has multiple handlers
    | #( "trap"
         {
	   TrapStatement *ts = create_trap_statement(suif_env);
	   Context nc = *c;
	   nc.traps = ts->get_symbol_table();
	   nc.traps->set_explicit_super_scope(c->traps);
	   nc.definitions = ts->get_definition_block();
	 }
         #(TRAPS ( trapDecl[&nc] )+ )
	 { Statement *s; }
         s=statement[&nc]
         {
	   ts->set_body(s);
	   Expression *e;
	 }
         ( #(htok:"handle" e=trapExpression[&nc] s=statement[&nc]
	   {
	     PredicatedStatement *ps =
	       create_predicated_statement(suif_env, s, e);
	     attach_token_linenumber(ps, htok);
	     ts->append_handler(ps);
	   }	
           )
         )*
         { st = ts; }
       )

    | #( "exit" etrap:ID { Expression *e = 0; } ( e=texpr:expression[c] )?
      {
        TrapSymbol *ts = find_trap(c, etrap);
	st = create_exit_statement( suif_env, ts, e);
	if (e) {
	  if ( ts->get_type() == info_block->get_q_void_type() ) {
	    report_error(etrap, "attempting to return value with pure trap `%s'",
			 ts->get_name().c_str());
	  } else if ( e->get_result_type() !=
		      ts->get_type()->get_base_type() ) {
	    report_error(etrap,
			 "returning wrong type for trap `%s'",
			 ts->get_name().c_str());
	    report_warning(ts, "trap of type %s",
			   ts->get_type()->get_name().c_str());
	    report_warning(texpr, "expression of type %s",
			   e->get_result_type()->get_name().c_str());
	  }
	} else {
	  // No expression
	  if ( ts->get_type() != info_block->get_q_void_type() ) {
	    report_error(etrap,
			 "valued trap `%s' requires exit with an expression",
			 ts->get_name().c_str());
	    report_warning(ts, "trap has type `%s'",
			   ts->get_type()->get_name().c_str());
	  }
	}
      }
      )

    | #("exec"
        { ExecStatement *es = create_exec_statement(suif_env); }
	(taskCall[c, es])+
        { st = es; }
       )

    | #( "var" {
         VariableStatement *vst = create_variable_statement(suif_env);
	 Context nc = *c;
	 nc.variables = vst->get_symbol_table();
	 nc.variables->set_explicit_super_scope(c->variables);	 
	 nc.definitions = vst->get_definition_block();
         QualifiedType *qt;
      }
      #(VARS
        ( #(VDECL
            ( variableName:ID
	      { Expression *e = 0; }
	      ( #(COLEQUALS e=varexpr:expression[c]) )?
	      qt=type[c]
	      { create_local_variable(&nc,vst,variableName,e,varexpr,qt); }
	    | #(ids:COMMA (ID)+) qt=type[c]
	        {
		  for ( RefLineAST id = #ids->getFirstChild() ; id ;
			id = id->getNextSibling() )
		    create_local_variable(&nc,vst,id,0,0,qt);
		}	
            )
          )
        )+
       )
       { Statement *s; }
       s=statement[&nc]
       {
	 vst->set_body(s);
	 st = vst;
       }
      )

    | #( "signal"
        {
	  SignalStatement *ss = create_signal_statement(suif_env);
	  Context nc = *c;
	  nc.signals = ss->get_symbol_table();
	  nc.signals->set_explicit_super_scope(c->signals);
	  nc.signal_vars = nc.signals;
	  nc.definitions = ss->get_definition_block();
        }
      #(SIGS ( signalDecl[&nc, localAnnote, false] )+ )
     { Statement *s; }
     s=statement[&nc]
     {
       ss->set_body(s);
       st = ss;
     }
    )

    | #(RUN module:ID
        {
	  RunStatement *rs =
	    create_run_statement(suif_env, module->getText().c_str(),
				 module->getText().c_str());
	}
        ( oldMod:ID { rs->set_old_module_name(oldMod->getText().c_str()); } )?
        ( typeRenaming[c,rs]
        | constantRenaming[c,rs]
        | functionRenaming[c,rs]
        | procedureRenaming[c,rs]
        | taskRenaming[c,rs]
        | signalRenaming[c,rs]
        )*
	{ st = rs; }
      )

    | #(DOWATCHING { Expression *e; Statement *s1, *s2 = 0; }
        s1=statement[c]
        e=de:delayExpression[c]
        (s2=statement[c])?
        { st = create_do_watching_statement(suif_env, s1, e, s2); }
      )
    | #(DOUPTO { Expression *e; Statement *s; }
        s=statement[c]
        e=delayExpression[c]
        { st = create_do_upto_statement(suif_env, s, e); }
    )
    { attach_token_linenumber(st,#statement); }
    )
  ;


expression [Context *c] returns [Expression *e]
  : { Expression *e1, *e2; }
    (
      #("or" e1=expression[c] e2=expression[c] )
      { e = build_boolean_binary( #expression, logical_or_op, e1, e2 ); }
    | #("and" e1=expression[c] e2=expression[c] )
      { e = build_boolean_binary( #expression, logical_and_op, e1, e2 ); }

    | #(PLUS e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, add_op, e1, e2 ); }
    | #(DASH e1=expression[c] (
           e2=expression[c]
	   { e = build_numeric_binary( #expression, subtract_op, e1, e2 ); }
         | /* nothing */ { e = build_numeric_unary( #expression, negate_op, e1 ); }
         ))

    | #(STAR e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, multiply_op, e1, e2 ); }
    | #(SLASH e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, divide_op, e1, e2 ); }
    | #("mod" e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, remainder_op, e1, e2 ); }

    | #(EQUALS e1=expression[c] e2=expression[c] )
      { e = build_equality_binary( #expression, is_equal_to_op, e1, e2 ); }
    | #(NEQUAL e1=expression[c] e2=expression[c] )
      { e = build_equality_binary( #expression, is_not_equal_to_op, e1, e2 ); }

    | #(LESSTHAN e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, is_less_than_op, e1, e2 ); }
    | #(GREATERTHAN e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, is_greater_than_op, e1, e2 ); }
    | #(LEQUAL e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, is_less_than_or_equal_to_op, e1, e2 ); }
    | #(GEQUAL e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, is_greater_than_or_equal_to_op, e1, e2 ); }

    | #("not" e1=expression[c])
      { e = build_boolean_unary( #expression, logical_not_op, e1 ); }

    | Integer {
        e = create_int_constant( suif_env, info_block->get_integer_type(),
				 #expression->getText().c_str());	
      }

    | FloatConst {
        e = create_float_constant( suif_env, info_block->get_float_type(),
                                   #expression->getText().c_str());
      }

    | DoubleConst {
        e = create_float_constant( suif_env, info_block->get_double_type(),
                                   #expression->getText().c_str());
      }
     
    | "true" {
        e = create_int_constant( suif_env, info_block->get_boolean_type(), 1);
      }

    | "false" {
        e = create_int_constant( suif_env, info_block->get_boolean_type(), 0);
      }

    | var:ID
      {
        VariableSymbol *vs = find_variable(c, var);
	DataType *dt = vs->get_type()->get_base_type();
        e = create_load_variable_expression( suif_env, dt, vs );
      }
    | str:StringConstant
      {
        e = create_string_constant(suif_env, info_block->get_string_type(),
				   str->getText().c_str());
      }
    | #(DQUESTION trap:ID)
      {
        TrapSymbol *ts = find_trap(c, trap);
	DataType *dt = ts->get_type()->get_base_type();
	if (dt == info_block->get_void_type()) {
	  report_error(trap,
		       "attempting to take value of pure exception `%s'",
		       trap->getText().c_str());
	}
	e = create_load_trap_value_expression(suif_env, dt, ts);
      }
    | #(QUESTION sig:ID)
      {
	SignalSymbol *ss = find_signal(c, sig);
	DataType *dt = ss->get_type()->get_base_type();
	if (dt == info_block->get_void_type()) {
	  report_error(sig,
		       "attempting to take value of pure signal `%s'",
		       sig->getText().c_str());
	}
	e = create_load_signal_value_expression(suif_env, dt, ss);
      }
    | #(CALL func:ID
        {
	  LString name = func->getText().c_str();
	  SymbolTableObject *sto = c->functions->lookup_lookup_table(name);
	  if (!sto) {
	    report_error(func, "undeclared function `%s'", name.c_str());

	    // Create an integer placeholder and skip the arguments
	    e = create_int_constant(suif_env, info_block->get_integer_type(),
				    0);
	    break;
	  }
	  EsterelFunctionSymbol *efs = to<EsterelFunctionSymbol>(sto);
	  EsterelFunctionType *eft = efs->get_type();
	  DataType *dt = eft->get_result()->get_base_type();
	  FunctionCall *fc = create_function_call(suif_env, dt, efs);
	  Expression *arg;
        }
        ( arg=expression[c] { fc->append_argument(arg); } )*
        {
	  if ( fc->get_argument_count() != eft->get_argument_count() ) {
	    report_error(func,
	      "wrong number of arguments to call of `%s' (expected %d got %d)",
			 name.c_str(), eft->get_argument_count(),
			 fc->get_argument_count());
	  }
	  for ( size_t i = 0 ;
		i < fc->get_argument_count() && i < eft->get_argument_count() ;
		i++ ) {
	    DataType *dt1 = fc->get_argument(i)->get_result_type();
	    DataType *dt2 = eft->get_argument(i)->get_base_type();
	    if ( dt1 != dt2 ) {
	      report_error(fc->get_argument(i),
			   "type of argument %d does not match (expected `%s' got `%s')",
			   i, dt2->get_name().c_str(), dt1->get_name().c_str());
	    }
	  }
	  e = fc;
        }
      )
    // FIXME: need "pre" operator
    )
    { attach_token_linenumber(e,#expression); }
  ;

sigExpression [Context *c]
  returns [Expression *e]
  : { Expression *e1, *e2; }
    (
      #( "and" e1=sigExpression[c] e2=sigExpression[c]
        { e = create_binary_expression( suif_env, e1->get_result_type(),
					"logical_and", e1, e2); } )
    | #( "or" e1=sigExpression[c] e2=sigExpression[c]
        { e = create_binary_expression( suif_env, e1->get_result_type(),
					"logical_or", e1, e2); } )
    | #( "not" e1=sigExpression[c]
        { e = create_unary_expression( suif_env, e1->get_result_type(),
					"logical_not", e1); } )
    | #( "pre" e1=sigExpression[c]
        { e = create_unary_expression( suif_env, e1->get_result_type(),
					"pre", e1); } )
    | sig:ID
      {
	SignalSymbol *ss = find_signal(c, sig);
	e = create_load_signal_expression(suif_env, info_block->get_boolean_type(), ss);
      }
    )
    { attach_token_linenumber(e,#sigExpression); }
  ;

delayExpression [Context *c]
  returns [Expression *e]
  : { Expression *e1, *e2; }
    (
      e=sigExpression[c]
    | #("immediate" e1=sigExpression[c]
      { e = create_delay(suif_env, info_block->get_boolean_type(),
			 e1, 0, true); }
      )
    | #(DELAY e2=count:expression[c] e1=sigExpression[c]
      {
        Delay * d = create_delay(suif_env, info_block->get_boolean_type(),
				 e1, e2);
        if ( e2->get_result_type() != info_block->get_integer_type() ) {
          report_error(count, "type of count expression must be integer, found `%s'",
		       e2->get_result_type()->get_name().c_str());
        }
        e = d;
      }
      )
    )
    { attach_token_linenumber(e,#delayExpression); }
  ;

trapExpression [Context *c] returns [Expression *e]
  : { Expression *e1, *e2;}
    ( trap:ID
      {
        TrapSymbol *ts = find_trap(c, trap);
        e = create_load_trap_expression(suif_env,
					info_block->get_boolean_type(), ts);
      }
    | #( "and" e1=trapExpression[c] e2=trapExpression[c]
        { e = create_binary_expression( suif_env, info_block->get_boolean_type(),
					"logical_and", e1, e2 ); } )
    | #( "or" e1=trapExpression[c] e2=trapExpression[c]
        { e = create_binary_expression( suif_env, info_block->get_boolean_type(),
					"logical_or", e1, e2 ); } )
    | #( "not" e1=trapExpression[c]
         { e = create_unary_expression( suif_env, info_block->get_boolean_type(), "logical_not", e1 ); } )
    )
  ;

procArgs [Context *c, ProcedureCallStatement *pcs]
  : #( VARS
      ( var:ID { pcs->append_reference_arg(find_variable(c, var)); } )*
    )
    #( ARGS { Expression *e; }
       ( e=expression[c] { pcs->append_value_arg(e); } )*
    )
    // FIXME: Typecheck this bad boy
  ;

taskCall [Context *c, ExecStatement *es]
  : #( cas:CASE task:ID
       {
	 LString name = task->getText().c_str();
	 suif_assert(c->tasks);
	 SymbolTableObject *sto = c->tasks->lookup_lookup_table(name);
	 if (!sto) {
	   report_error(task, "undeclared task `%s'", name.c_str());
	   return;
	 }
	 suif_assert(sto);
	 EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	 suif_assert(eps);
	 EsterelProcedureType *ept = eps->get_type();
	 TaskCall *tc = create_task_call(suif_env, eps, 0);
	 es->append_call(tc);
	 attach_token_linenumber(tc, cas);
	 suif_assert(ept);
       }
       procArgs[c, tc]
       sig:ID
       {
	 SignalSymbol *ss = find_signal(c, sig);
	 tc->set_signal(ss);
	 if ( ss->get_direction() != returnAnnote ) {
	   report_error(sig, "%s signal `%s' not of type `return'",
			ss->get_direction().c_str(),
			ss->get_name().c_str());
	   report_warning(ss, "`%s' declared here", ss->get_name().c_str() );
	 }
	 Statement *s;
       }
       ( s=statement[c] { tc->set_body(s); } )?
     )
  ;

typeRenaming [Context *c, RunStatement *rs]
 : #( TRENAME { QualifiedType *qt; } qt=newType:type[c] oldID:ID
     {
       TypeRenaming *tr =
	 create_type_renaming(suif_env, oldID->getText().c_str(), qt);
       attach_token_linenumber(tr, newType);
       rs->append_type_renaming(tr);
     }
   )
 ;    

constantRenaming [Context *c, RunStatement *rs]
 : #( CRENAME { Expression *e; } e=expression[c] oldID:ID
     {
       ConstantRenaming *cr =
	 create_constant_renaming(suif_env, oldID->getText().c_str(), e);
       attach_token_linenumber(cr, #constantRenaming);
       rs->append_constant_renaming(cr);
     }
   )
 ;    

functionRenaming [Context *c, RunStatement *rs]
 : #( FRENAME { EsterelFunctionSymbol *efs; }
     efs=func:newFunctionName[c]
     oldID:ID
     {
       if (efs) {
	 FunctionRenaming *fr =
	   create_function_renaming(suif_env, oldID->getText().c_str(), efs);
	 attach_token_linenumber(fr, func);
	 suif_assert(rs);
	 rs->append_function_renaming(fr);
       }
     }
   )
 ;

newFunctionName [Context *c] returns [EsterelFunctionSymbol *efs]
  : { efs = 0; }
    ( func:ID
       {
	 LString name = func->getText().c_str();
	 SymbolTableObject *sto = c->functions->lookup_lookup_table(name);
	 if (!sto) {
	   report_error(func, "undeclared function `%s'", name.c_str());
	 } else {
	   efs = to<EsterelFunctionSymbol>(sto);
	 }
       }
    | pf:predefinedFunction
      { efs = get_builtin_function(c, pf->getText().c_str()); }
    )
  ;

predefinedFunction
  : "and"
  | "or"
  | "not"
  | PLUS
  | DASH
  | STAR
  | SLASH
  | "mod"
  | LESSTHAN
  | GREATERTHAN
  | LEQUAL
  | GEQUAL
  | NEQUAL
  | EQUALS
  ;

procedureRenaming [Context *c, RunStatement *rs]
 : #( PRENAME proc:ID oldID:ID
     {
       LString name = proc->getText().c_str();
       SymbolTableObject *sto = c->procedures->lookup_lookup_table(name);
       if (!sto) {
	 report_error(proc, "undeclared procedure `%s'", name.c_str());
       } else {
	 EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	 ProcedureRenaming *pr =
	   create_procedure_renaming(suif_env, oldID->getText().c_str(), eps);
	 attach_token_linenumber(pr, proc);
	 rs->append_procedure_renaming(pr);
       }
     }
   )
 ;

taskRenaming [Context *c, RunStatement *rs]
 : #( TARENAME proc:ID oldID:ID
     {
       LString name = proc->getText().c_str();
       SymbolTableObject *sto = c->tasks->lookup_lookup_table(name);
       if (!sto) {
	 report_error(proc, "undeclared task `%s'", name.c_str());
       } else {
	 EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	 ProcedureRenaming *tr =
	   create_procedure_renaming(suif_env, oldID->getText().c_str(), eps);
	 attach_token_linenumber(tr, proc);
	 rs->append_task_renaming(tr);
       }
     }
   )
 ;

signalRenaming [Context *c, RunStatement *rs]
 : #( SRENAME sig:ID oldID:ID
     {
       SignalSymbol *ss = find_signal(c, sig);
       SignalRenaming *sr =
	   create_signal_renaming(suif_env, oldID->getText().c_str(), ss);
       attach_token_linenumber(sr, sig);
       rs->append_signal_renaming(sr);
     }
   )
 ;
